#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:zhoukaiyin
"""

import os
import importlib
# automatically import any Python files in the layers/ directory
for file in os.listdir(os.path.dirname(__file__)):
    if file.endswith(".py") and not file.startswith("_"):
        dataset_name = file[: file.find(".py")]
        module = importlib.import_module("fennlp.layers." + dataset_name)


#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:zhoukaiyin
@Tensorflow 2.0
All of the following Code was follow Google BERT!
"""
import tensorflow as tf
from fennlp.tools import create_initializer, get_shape_list


class WDEmbedding(tf.keras.layers.Layer):
    def __init__(self,
                 vocab_size,
                 embedding_size=128,
                 initializer_range=0.02,
                 word_embedding_name="word_embeddings",
                 use_one_hot_embedding=False,
                 name=None,
                 **kwargs):
        super(WDEmbedding, self).__init__(name=name, **kwargs)
        self.vocab_size = vocab_size
        self.embedding_size = embedding_size
        self.initializer_range = initializer_range
        self.word_embedding_name = word_embedding_name
        self.use_one_hot_embedding = use_one_hot_embedding

    def build(self, input_shape):
        self.input_spec = tf.keras.layers.InputSpec(shape=input_shape)
        self.embedding_table = self.add_weight(
            name=self.word_embedding_name,
            dtype=tf.keras.backend.floatx(),
            shape=[self.vocab_size, self.embedding_size],
            initializer=create_initializer(self.initializer_range),
            trainable=True,
        )
        self.built = True

    def call(self, input_ids):
        if input_ids.shape.ndims == 2:
            input_ids = tf.expand_dims(input_ids, axis=[-1])
        flat_input_ids = tf.reshape(input_ids, [-1])
        if self.use_one_hot_embedding:
            one_hot_input_ids = tf.keras.backend.one_hot(flat_input_ids, self.vocab_size)
            output = tf.linalg.matmul(one_hot_input_ids, self.embedding_table)
        else:
            output = tf.gather(self.embedding_table, flat_input_ids)
        input_shape = get_shape_list(input_ids)
        output = tf.reshape(output, input_shape[0:-1] + [input_shape[-1] * self.embedding_size])
        return output


class SegPosEmbedding(tf.keras.layers.Layer):
    def __init__(self,
                 use_token_type=True,
                 hidden_dropout_prob=0.01,
                 token_type_vocab_size=2,
                 token_type_embedding_name="token_type_embeddings",
                 use_position_embeddings=True,
                 position_embedding_name="position_embeddings",
                 initializer_range=0.02,
                 max_position_embeddings=512,
                 name=None,
                 **kwargs):

        super(SegPosEmbedding, self).__init__(name=name, **kwargs)
        self.use_token_type = use_token_type
        self.token_type_vocab_size = token_type_vocab_size
        self.token_type_embedding_name = token_type_embedding_name
        self.use_position_embeddings = use_position_embeddings
        self.position_embedding_name = position_embedding_name
        self.initializer_range = initializer_range
        self.max_position_embeddings = max_position_embeddings
        self.hidden_dropout_prob = hidden_dropout_prob

    def build(self, input_shape):

        input_ids_shape = input_shape
        self.input_spec = tf.keras.layers.InputSpec(shape=input_ids_shape)

        self.token_type_table = self.add_weight(
            name=self.token_type_embedding_name,
            shape=[self.token_type_vocab_size, input_shape[2]],
            dtype=tf.keras.backend.floatx(),
            initializer=create_initializer(self.initializer_range),
            trainable=True
        )

        self.full_position_embeddings = self.add_weight(
            name=self.position_embedding_name,
            shape=[self.max_position_embeddings, input_shape[2]],
            dtype=tf.keras.backend.floatx(),
            initializer=create_initializer(self.initializer_range),
            trainable=True)
        self.drop_out = tf.keras.layers.Dropout(self.hidden_dropout_prob)
        self.layer_norm = tf.keras.layers.LayerNormalization(axis=-1, name="LayerNorm")

        self.built = True

    def call(self, input_tensor, token_type_ids=None,is_training=True):
        inputshape = get_shape_list(input_tensor, expected_rank=3)
        batch_size = inputshape[0]
        seq_length = inputshape[1]
        width = inputshape[2]
        output = input_tensor
        # segment features
        if self.use_token_type:
            if token_type_ids is None:
                raise ValueError("token_type_ids must be specified if use_token_type is True")
            flat_token_type_ids = tf.reshape(token_type_ids, [-1])
            one_hot_ids = tf.one_hot(flat_token_type_ids, depth=self.token_type_vocab_size)
            token_type_embeddings = tf.linalg.matmul(one_hot_ids, self.token_type_table)
            token_type_embeddings = tf.reshape(token_type_embeddings, [batch_size, seq_length, width])
            output += token_type_embeddings
        # position features
        if self.use_position_embeddings:
            position_embeddings = tf.slice(self.full_position_embeddings, [0, 0], [seq_length, -1])
            # num_dims = len(output.shape.as_list())
            num_dims = len(output.shape.as_list())
            position_broadcast_shape = []
            for _ in range(num_dims - 2):
                position_broadcast_shape.append(1)
            position_broadcast_shape.extend([seq_length, width])
            position_embeddings = tf.reshape(position_embeddings, position_broadcast_shape)
            output += position_embeddings
        output = self.layer_norm(output)
        # in official work they not use training
        output = self.drop_out(output, training=is_training)
        return output



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Types for typing functions signatures."""

from typing import Union, Callable, List

import numpy as np
import tensorflow as tf


Number = Union[
    float,
    int,
    np.float16,
    np.float32,
    np.float64,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
]

Initializer = Union[None, dict, str, Callable]
Regularizer = Union[None, dict, str, Callable]
Constraint = Union[None, dict, str, Callable]
Activation = Union[None, str, Callable]

TensorLike = Union[
    List[Union[Number, list]],
    tuple,
    Number,
    np.ndarray,
    tf.Tensor,
    tf.SparseTensor,
    tf.Variable,
]
FloatTensorLike = Union[tf.Tensor, float, np.float16, np.float32, np.float64]
AcceptableDTypes = Union[tf.DType, np.dtype, type, int, str, None]


#!encoding:utf-8
import requests
import sys
import zipfile
import os
import json


class LoadCheckpoint(object):
    def __init__(self, langurage='zh'):
        self.lg = langurage
        if langurage == "zh":

            url = "https://storage.googleapis.com/bert_models/" \
                  "2018_11_03/chinese_L-12_H-768_A-12.zip"
        elif langurage == "en":
            url = "https://storage.googleapis.com/bert_models/" \
                  "2018_10_18/uncased_L-12_H-768_A-12.zip"
        self.url = url
        self.size = self.getsize()
        self.unzip()

    def getsize(self):
        try:
            r = requests.head(self.url)
            size = r.headers.get("Content-Length")
            return int(size)
        except:
            print("Failed Download!")
            exit()

    def bar(self, num, total):
        rate = num / total
        rate_num = int(rate * 100)
        if rate_num == 100:
            r = '\r%s>%d%%\n' % ('=' * int(rate_num / 3), rate_num,)  # 控制等号输出数量，除以3,表示显示1/3
        else:
            r = '\r%s>%d%%' % ('=' * int(rate_num / 3), rate_num,)
        sys.stdout.write(r)
        sys.stdout.flush()

    def download(self, chunk_size=1024):
        num = 0
        response = requests.get(self.url, stream=True)
        with open(self.url.split('/')[-1], 'wb') as wf:
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:
                    wf.write(chunk)
                    num += chunk_size
                    self.bar(num, self.size)

    def unzip(self):
        filename = self.url.split('/')[-1]
        if not os.path.exists(filename):
            open(filename, 'w').close()
        if os.path.getsize(filename) != self.size:
            print("Download and unzip: {}".format(filename))
            self.download()
        zip_file = zipfile.ZipFile(filename)
        zip_file.extractall()

    def load_bert_param(self, pretraining=False):
        filename = self.url.split('/')[-1]
        config = "{}/{}".format(filename.split('.')[0], "bert_config.json")
        vocab_file = "{}/{}".format(filename.split('.')[0], "vocab.txt")
        model_path = "{}/{}".format(filename.split('.')[0], "bert_model.ckpt")
        bert_param = json.load(open(config, 'r'))
        if not pretraining and self.lg == 'zh':
            bert_param.pop("directionality")
            bert_param.pop("pooler_fc_size")
            bert_param.pop("pooler_num_attention_heads")
            bert_param.pop("pooler_num_fc_layers")
            bert_param.pop("pooler_size_per_head")
            bert_param.pop("pooler_type")
        if not pretraining and self.lg == 'en':
            pass
        bert_param["batch_size"] = 32
        bert_param["maxlen"] = 80
        bert_param["label_size"] = 10
        return bert_param, vocab_file, model_path



#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:Kaiyin Zhou
"""
import tensorflow as tf
import numpy as np


def broadcast(src, other, dim: int):
    if dim < 0:
        dim = len(other.shape) + dim
    if len(src.shape) == 1:
        for _ in range(0, dim):
            src = tf.expand_dims(src, 0)
    for _ in range(len(src.shape), len(other.shape)):
        src = tf.expand_dims(src, -1)

    # src = src.expand_as(other)
    shape = other.get_shape().as_list()
    shape[0] = 1
    src = tf.tile(src, shape)
    return tf.cast(src, tf.int64)


def scatter_sum(src, index, dim: int = -1,
                out=None,
                dim_size=None):
    index = broadcast(index, src, dim)
    if out is None:
        size = src.get_shape().as_list()
        if dim_size is not None:
            size[dim] = dim_size
        elif index.numel() == 0:
            size[dim] = 0
        else:
            size[dim] = int(index.max()) + 1
        out = tf.zeros(size, dtype=src.dtype)
        print(out)
        print(index)
        print(src)
        return tf.tensor_scatter_nd_add(out, index, src)
    else:
        return tf.tensor_scatter_nd_add(out, index, src)





#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:zhoukaiyin
"""
import tensorflow as tf
import os
import codecs
import numpy as np
from tqdm import tqdm
from gensim.models import KeyedVectors

def get_embeddings(word,*arr):
    return word,np.array(arr,dtype=np.float32)

def build_matrix(embeddings_index,vocab_index,emb_dim):
    vocab_size = len(vocab_index)
    embeddings_matrix = np.zeros((len(vocab_index), emb_dim))
    for word, i in tqdm(vocab_index.items()):
        if i >= vocab_size: continue
        try:
            embeding_vector = embeddings_index[word]
        except:
            embeding_vector = embeddings_index["unknown"]
        if embeding_vector is not None:
            embeddings_matrix[i] = embeding_vector
    return embeddings_matrix

def glove(vocab_index_path,emb_dim):
    """
    get all vocabs' embedding
    :param vocab:
    :param emb_dim:
    :return:
    """
    # lf = codecs.open(vocab_index_path)
    # vocab_index = pickle.load(lf)
    vocab_index = np.load(vocab_index_path).item()
    url ="./embedding/glove.6B.{}d.txt".format(emb_dim)
    glove_name= "glove.6B.{}d.txt".format(emb_dim)
    try:
        data = tf.keras.utils.get_file("glove",origin=url)
        dirname = os.path.dirname(data)
    except ValueError:
        dirname = "./embedding/"

    embeddings_index = dict(get_embeddings(*o.split(" ")) for o in
                            codecs.open(os.path.join(dirname,glove_name),
                                        encoding='utf-8'))
    return build_matrix(embeddings_index, vocab_index, emb_dim)

def fast_text(vocab_index_path,emb_dim):
    vocab_index = np.load(vocab_index_path)

    url = "./embedding/crawl-{}d-2M.vec".format(emb_dim)
    fasttext_name = "crawl-{}d-2M.vec".format(emb_dim)
    try:
        data = tf.keras.utils.get_file("fasttext", origin=url)
        dirname = os.path.dirname(data)
    except ValueError:
        dirname = "./embedding/"

    embeddings_index = dict(
        get_embeddings(*o.strip().split(" ")) for o in codecs.open(os.path.join(dirname,fasttext_name),
                                        encoding='utf-8') if len(o) > 100)
    fasttext_embedding_matrix = build_matrix(embeddings_index, vocab_index,emb_dim)
    return fasttext_embedding_matrix

def word2vec(vocab_index_path,emb_dim):
    vocab_index = np.load(vocab_index_path)

    url = "./embedding/word2vec{}d.bin.gz".format(emb_dim)
    word2vec_name = "word2vec{}d.bin.gz".format(emb_dim)
    try:
        data = tf.keras.utils.get_file("word2vec", origin=url)
        dirname = os.path.dirname(data)
    except ValueError:
        dirname = "./embedding/"
    word2vec = KeyedVectors.load_word2vec_format(os.path.join(dirname,word2vec_name), binary=True)
    embeddings_index = {}
    for i, vec in tqdm(enumerate(word2vec.wv.vectors)):
        embeddings_index[word2vec.wv.index2word[i]] = vec
    word2vec_embedding_matrix = build_matrix(embeddings_index,  vocab_index,emb_dim)
    del word2vec
    return word2vec_embedding_matrix




#! encoding="utf-8"
import tensorflow as tf


class RelationalGraphConvolution(tf.keras.layers.Layer):
    def __init__(self, output_dim,
                 num_relations,
                 num_bases,
                 activation='linear',
                 kernel_initializer=None,
                 bias_initializer=None,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 use_bias=True,
                 **kwargs):
        super(RelationalGraphConvolution, self).__init__(self, **kwargs)

        self.kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)
        self.kernel_regularizer = tf.keras.regularizers.get(kernel_regularizer)
        self.bias_regularizer = tf.keras.regularizers.get(bias_regularizer)
        self.activation = tf.keras.activations.get(activation)
        self.use_bias = use_bias
        self.output_dim = output_dim
        self.num_relations = num_relations
        self.num_bases = num_bases

    def build(self, input_shape):
        self.in_features = input_shape[1]
        # V_b   [B,in_dim,out_dim]
        self.basis = self.add_weight(
            shape=(self.num_bases, self.in_features, self.output_dim),
            name='basis',
        )

        self.att = self.add_weight(
            shape=(self.num_relations, self.num_bases),
            name='kernel',
        )
        # W_0^l
        self.weight = self.add_weight(
            shape=(self.in_features, self.output_dim),
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            name='kernel',
        )
        # bias
        if self.use_bias:
            self.bias = self.add_weight(
                shape=(self.output_dim,),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                name='bias',
            )

    def message(self, x_j, edge_index_j, edge_type, edge_norm):
        # wr [R,in_dim*out_dim]
        wr = tf.linalg.matmul(self.att, tf.reshape(self.basis, (self.num_bases, -1)))

        if x_j is None:
            # wr [R*in_dm, out_dim]
            wr = tf.reshape(wr, (-1, self.output_dim))
            index = edge_type * self.in_features + edge_index_j
            output = tf.gather(wr, index)
        else:
            wr = tf.reshape(wr, (self.num_relations, -1, self.output_dim))
            wr = tf.gather(wr, tf.cast(edge_type, tf.int64))
            print(wr)
            output = tf.linalg.matmul(tf.expand_dims(x_j, 1), wr)
            output = tf.squeeze(output, -2)
        return output if edge_norm is None else output * tf.reshape(edge_norm, (-1, 1))

    def aggregate(self, outputs):
        return tf.reduce_mean(outputs)

    def call(self, x, edge_index, edge_type, edge_norm):
        """
        :param x: entity
        :param edge_index: edge_index
        :param edge_type: edge_type
        :param edge_norm:edge_norm
        :return:
        """
        output = self.message(x, edge_index, edge_type, edge_norm)
        # TODO
        aggr_out = self.aggregate(output)
        if x is None:
            output = aggr_out + self.weight
        else:
            output = aggr_out + tf.linalg.matmul(x, self.weight)
        if self.use_biase:
            output += self.bias
        return output



#! encoding:utf-8

import tensorflow as tf
import numpy as np
import os
from scipy import sparse
from collections import defaultdict


class TuckERLoader():
    def __init__(self, base_path="data", reverse=True):
        self.train_data = self.load_data(base_path, 'train.txt', reverse=reverse)
        self.valid_data = self.load_data(base_path, 'valid.txt', reverse=reverse)
        self.test_data = self.load_data(base_path, 'test.txt', reverse=reverse)
        self.data = self.train_data + self.valid_data + self.test_data
        self.entities = self.get_entities(self.data)

        self.train_relations = self.get_relations(self.train_data)
        self.valid_relations = self.get_relations(self.valid_data)
        self.test_relations = self.get_relations(self.test_data)

        self.relations = self.train_relations + \
                         [i for i in self.valid_relations if i not in self.train_relations] + \
                         [i for i in self.test_relations if i not in self.train_relations]

        self.entity_idxs = {self.entities[i]: i for i in range(len(self.entities))}
        self.realtion_idxs = {self.relations[i]: i for i in range(len(self.relations))}

        del self.train_relations
        del self.test_relations
        del self.valid_relations

    def data_dump(self, data="train"):
        if data == "train":
            data = self.train_data
        elif data == "valid":
            data = self.valid_data
        elif data == "test":
            data = self.test_data
        data_idxs = self.get_data_idxs(data)
        er_vocab = self.get_er_vocab(data_idxs)
        er_vocab_pairs = list(er_vocab.keys())
        return er_vocab, er_vocab_pairs

    def load_data(self, base_path, data_type="train", reverse=False):
        data = []
        with open(os.path.join(base_path, data_type)) as rf:
            for line in rf:
                contents = line.strip().split()
                data.append(contents)
                if reverse:
                    data.append([contents[2], contents[1] + "_reverse", contents[0]])
            return data

    def get_relations(self, data):
        relations = sorted(list(set([d[1] for d in data])))
        return relations

    def get_entities(self, data):
        entities = sorted(list(set([d[0] for d in data] + [d[2] for d in data])))
        return entities

    def get_batch(self, er_vocab, er_vocab_pairs, batch_size=32):

        #
        targets = [er_vocab[key] for key in er_vocab_pairs]

        def generator():
            for x, y in zip(er_vocab_pairs, targets):
                yield {'h_r': x, 't': y}

        dataset = tf.data.Dataset.from_generator(
            generator=generator,
            output_types={'h_r': tf.int32, 't': tf.int32})
        dataset = dataset.padded_batch(batch_size,
                                       padded_shapes={'h_r': [None], 't': [None]},
                                       drop_remainder=True)
        return dataset

    def get_data_idxs(self, data):  # data could be self.train_data self.valid_data self.test_data
        data_idxs = [
            (self.entity_idxs[data[i][0]], self.realtion_idxs[data[i][1]],
             self.entity_idxs[data[i][2]]) for i in range(len(data))
        ]
        print("Number of data points: %d" % len(data_idxs))
        return data_idxs

    def get_er_vocab(self, data_index):
        er_vocab = defaultdict(list)
        for triple in data_index:
            er_vocab[(triple[0], triple[1])].append(triple[2])
        return er_vocab

    def target_convert(self, targets, batch_size, num_entities):
        targets_one_hot = np.zeros((batch_size, num_entities))
        for idx, tar in enumerate(targets):
            targets_one_hot[idx, tf.gather_nd(tar, tf.where(tar > 0))] = 1
        return tf.constant(targets_one_hot)


class GCNLoader():
    def __init__(self, base_path="data", dataset="cora"):
        self.base_path = base_path
        self.dataset = dataset
        print("Loading {} dataset...".format(dataset))

    def encode_onehot(self, labels):
        classes = set(labels)
        classes_dict = {c: np.identity(len(classes))[i, :] for i, c in
                        enumerate(classes)}
        labels_onehot = np.array(list(map(classes_dict.get, labels)),
                                 dtype=np.int32)
        return labels_onehot

    def normalize(self, mx):
        rowsum = np.array(mx.sum(1))
        r_inv = (1 / rowsum).flatten()
        r_inv[np.isinf(r_inv)] = 0.
        r_mat_inv = sparse.diags(r_inv)
        mx = r_mat_inv.dot(mx)
        return mx

    def convert_2_sparse_tensor(self, sparse_matrix):
        sparse_matrix = sparse_matrix.tocoo().astype(np.float32)
        values = sparse_matrix.data
        shape = sparse_matrix.shape
        indices = np.array([[row, col] for row, col in zip(sparse_matrix.row, sparse_matrix.col)], dtype=np.int64)
        return tf.sparse.SparseTensor(indices, values, shape)

    def load(self):
        idx_features_labels = np.genfromtxt("{}/{}.content".format(self.base_path, self.dataset),
                                            dtype=np.dtype(str))
        features = sparse.csr_matrix(idx_features_labels[:, 1:-1], dtype=np.float32)
        labels = self.encode_onehot(idx_features_labels[:, -1])

        # 构建图
        idx = np.array(idx_features_labels[:, 0], dtype=np.int32)
        idx_map = {j: i for i, j in enumerate(idx)}
        edges_unordered = np.genfromtxt("{}/{}.cites".format(self.base_path, self.dataset),
                                        dtype=np.int32)
        # [[1,2],
        #  [22,23]]
        # N*2
        edges = np.array(list(map(idx_map.get, edges_unordered.flatten())),
                         dtype=np.int32).reshape(edges_unordered.shape)
        adj = sparse.coo_matrix((np.ones(edges.shape[0]), (edges[:, 0], edges[:, 1])),
                                shape=(labels.shape[0], labels.shape[0]),
                                dtype=np.float32)

        # 构建对称邻接矩阵
        adj = adj + adj.T.multiply(adj.T > adj) - adj.multiply(adj.T > adj)

        features = self.normalize(features)
        adj = self.normalize(adj + sparse.eye(adj.shape[0]))

        features = tf.constant(np.array(features.todense()))

        labels = tf.constant(np.where(labels)[1])
        adj = self.convert_2_sparse_tensor(adj)

        idx_train = range(140)
        idx_val = range(200, 500)
        idx_test = range(500, 1500)

        return features, adj, labels, idx_train, idx_val, idx_test


class RGCNLoader(object):
    def __init__(self, base_path="data", dataset="FB15k"):
        self.base_path = base_path
        self.dataset = dataset
        self.file_path = os.path.join(base_path, dataset)

    def read_triplets(self, file_path, entity2id, relation2id):
        triplets = []

        with open(file_path) as f:
            for line in f:
                head, relation, tail = line.strip().split('\t')
                triplets.append((entity2id[head], relation2id[relation], entity2id[tail]))

        return np.array(triplets)

    def load_data(self):
        print("load data from {}".format(self.file_path))

        with open(os.path.join(self.file_path, 'entities.dict')) as f:
            entity2id = dict()

            for line in f:
                eid, entity = line.strip().split('\t')
                entity2id[entity] = int(eid)

        with open(os.path.join(self.file_path, 'relations.dict')) as f:
            relation2id = dict()

            for line in f:
                rid, relation = line.strip().split('\t')
                relation2id[relation] = int(rid)

        train_triplets = self.read_triplets(os.path.join(self.file_path, 'train.txt'), entity2id, relation2id)
        valid_triplets = self.read_triplets(os.path.join(self.file_path, 'valid.txt'), entity2id, relation2id)
        test_triplets = self.read_triplets(os.path.join(self.file_path, 'test.txt'), entity2id, relation2id)

        print('num_entity: {}'.format(len(entity2id)))
        print('num_relation: {}'.format(len(relation2id)))
        print('num_train_triples: {}'.format(len(train_triplets)))
        print('num_valid_triples: {}'.format(len(valid_triplets)))
        print('num_test_triples: {}'.format(len(test_triplets)))

        return entity2id, relation2id, train_triplets, valid_triplets, test_triplets

    def sample_edge_uniform(self, n_triples, sample_size):
        """Sample edges uniformly from all the edges."""
        all_edges = np.arange(n_triples)
        return np.random.choice(all_edges, sample_size, replace=False)

    def negative_sampling(self, pos_samples, num_entity, negative_rate):
        size_of_batch = len(pos_samples)
        num_to_generate = size_of_batch * negative_rate
        neg_samples = np.tile(pos_samples, (negative_rate, 1))
        labels = np.zeros(size_of_batch * (negative_rate + 1), dtype=np.float32)
        labels[: size_of_batch] = 1
        values = np.random.choice(num_entity, size=num_to_generate)
        choices = np.random.uniform(size=num_to_generate)
        subj = choices > 0.5
        obj = choices <= 0.5
        neg_samples[subj, 0] = values[subj]
        neg_samples[obj, 2] = values[obj]

        return np.concatenate((pos_samples, neg_samples)), labels

    def edge_normalization(self, edge_type, edge_index, num_entity, num_relation):
        from fennlp.gnn.scatter import scatter_sum
        '''
        
            Edge normalization trick
            - one_hot: (num_edge, num_relation)
            - deg: (num_node, num_relation)
            - index: (num_edge)
            - deg[edge_index[0]]: (num_edge, num_relation)
            - edge_norm: (num_edge)
        '''
        one_hot = tf.one_hot(tf.cast(edge_type, np.int32),
                             2 * num_relation, dtype=tf.int64)
        one_hot = tf.constant(one_hot.numpy())
        deg = scatter_sum(one_hot, edge_index[0], dim=0, dim_size=num_entity)
        index = edge_type + tf.keras.backend.arange(len(edge_index[0])) * (2 * num_relation)
        edge_norm = 1 / np.reshape(deg[edge_index[0]], -1)[index]

        return edge_norm

    def generate_sampled_graph_and_labels(self, triplets, batch_size, split_size, num_entity, num_rels, negative_rate):
        """
            Get training graph and signals
            First perform edge neighborhood sampling on graph, then perform negative
            sampling to generate negative samples
        """

        edges = self.sample_edge_uniform(len(triplets), batch_size)

        # Select sampled edges
        edges = triplets[edges]
        src, rel, dst = edges.transpose()
        uniq_entity, edges = np.unique((src, dst), return_inverse=True)
        src, dst = np.reshape(edges, (2, -1))
        relabeled_edges = np.stack((src, rel, dst)).transpose()

        # Negative sampling
        samples, labels = self.negative_sampling(relabeled_edges, len(uniq_entity), negative_rate)
        # samples 是所有的三元组，labels是表示该三元组是真是假
        # further split graph, only half of the edges will be used as graph
        # structure, while the rest half is used as unseen positive samples
        split_size = int(batch_size * split_size)
        graph_split_ids = np.random.choice(np.arange(batch_size),
                                           size=split_size, replace=False)

        src = tf.constant(src[graph_split_ids], dtype=tf.float32)
        dst = tf.constant(dst[graph_split_ids], dtype=tf.float32)
        rel = tf.constant(rel[graph_split_ids], dtype=tf.float32)
        # Create bi-directional graph
        src, dst = tf.concat((src, dst), axis=0), tf.concat((dst, src), axis=0)
        rel = tf.concat((rel, rel + num_rels), axis=0)
        edge_type = rel
        self.edge_index = tf.stack((src, dst))
        self.entity = tf.constant(uniq_entity)
        self.edge_type = edge_type
        self.edge_norm = tf.ones(edge_type.shape)

        self.samples = tf.constant(samples)
        self.labels = tf.constant(labels)



#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:Kaiyin Zhou
"""
import tensorflow as tf

class CrfLogLikelihood(tf.keras.layers.Layer):
    def __init__(self, name='crf', **kwargs):
        super(CrfLogLikelihood, self).__init__(name=name, **kwargs)

    def build(self, input_shape):
        num_tags = input_shape[2]
        initializer = tf.keras.initializers.GlorotUniform()
        self.transition_params = self.add_weight(
            shape=[num_tags, num_tags],
            initializer=initializer,
            name="transitions")
    def call(self, inputs, tag_indices, sequence_lengths):
        # cast type to handle different types
        tag_indices = tf.cast(tag_indices, dtype=tf.int32)
        sequence_lengths = tf.cast(sequence_lengths, dtype=tf.int32)
        sequence_scores = crf_sequence_score(inputs, tag_indices, sequence_lengths,
                                             self.transition_params)
        # log_sum(exp(一条路劲得分))，对所有路径得分的exp求sum后取log
        log_norm = crf_log_norm(inputs, sequence_lengths, self.transition_params)
        # Normalize the scores to get the log-likelihood per example.
        log_likelihood = sequence_scores - log_norm
        return log_likelihood, self.transition_params


def crf_sequence_score(inputs, tag_indices, sequence_lengths,
                       transition_params):
    tag_indices = tf.cast(tag_indices, dtype=tf.int32)
    sequence_lengths = tf.cast(sequence_lengths, dtype=tf.int32)
    unary_scores = crf_unary_score(tag_indices, sequence_lengths, inputs)
    # trans score
    binary_scores = crf_binary_score(tag_indices, sequence_lengths,
                                     transition_params)
    sequence_scores = unary_scores + binary_scores
    return sequence_scores

def crf_unary_score(tag_indices, sequence_lengths, inputs):
    tag_indices = tf.cast(tag_indices, dtype=tf.int32)
    sequence_lengths = tf.cast(sequence_lengths, dtype=tf.int32)

    batch_size = tf.shape(inputs)[0]
    max_seq_len = tf.shape(inputs)[1]
    num_tags = tf.shape(inputs)[2]

    flattened_inputs = tf.reshape(inputs, [-1])

    offsets = tf.expand_dims(tf.range(batch_size) * max_seq_len * num_tags, 1)  # 计算每一个batch起始位置
    offsets += tf.expand_dims(tf.range(max_seq_len) * num_tags, 0)  # 计算每一个step的预测结果的起始位置
    # Use int32 or int64 based on tag_indices' dtype.
    if tag_indices.dtype == tf.int64:
        offsets = tf.cast(offsets, tf.int64)
    flattened_tag_indices = tf.reshape(offsets + tag_indices, [-1])  # 这就获得每一个标签所在的index

    unary_scores = tf.reshape(
        tf.gather(flattened_inputs, flattened_tag_indices),
        [batch_size, max_seq_len])  # 将每一个标签所在的index取出来，获得模型预测得分

    masks = tf.sequence_mask(
        sequence_lengths, maxlen=tf.shape(tag_indices)[1], dtype=tf.float32)

    unary_scores = tf.reduce_sum(unary_scores * masks, 1)
    return unary_scores  # 获得模型预测得分

def crf_binary_score(tag_indices, sequence_lengths, transition_params):
    tag_indices = tf.cast(tag_indices, dtype=tf.int32)
    sequence_lengths = tf.cast(sequence_lengths, dtype=tf.int32)

    num_tags = tf.shape(transition_params)[0]
    num_transitions = tf.shape(tag_indices)[1] - 1

    # Truncate by one on each side of the sequence to get the start and end
    # indices of each transition.
    start_tag_indices = tf.slice(tag_indices, [0, 0], [-1, num_transitions])
    end_tag_indices = tf.slice(tag_indices, [0, 1], [-1, num_transitions])

    # Encode the indices in a flattened representation.
    # 太巧妙了
    flattened_transition_indices = start_tag_indices * \
                                   num_tags + end_tag_indices
    flattened_transition_params = tf.reshape(transition_params, [-1])

    # Get the binary scores based on the flattened representation.
    binary_scores = tf.gather(flattened_transition_params,
                              flattened_transition_indices)
    masks = tf.sequence_mask(
        sequence_lengths, maxlen=tf.shape(tag_indices)[1], dtype=tf.float32)
    truncated_masks = tf.slice(masks, [0, 1], [-1, -1])
    binary_scores = tf.reduce_sum(binary_scores * truncated_masks, 1)
    return binary_scores

def crf_log_norm(inputs, sequence_lengths, transition_params):

    sequence_lengths = tf.cast(sequence_lengths, dtype=tf.int32)
    # Split up the first and rest of the inputs in preparation for the forward
    # algorithm.
    first_input = tf.slice(inputs, [0, 0, 0], [-1, 1, -1])
    first_input = tf.squeeze(first_input, [1])
    """Forward computation of alpha values."""
    rest_of_input = tf.slice(inputs, [0, 1, 0], [-1, -1, -1])
    # Compute the alpha values in the forward algorithm in order to get the
    # partition function.

    alphas = crf_forward(rest_of_input, first_input, transition_params,
                         sequence_lengths)
    log_norm = tf.reduce_logsumexp(alphas, [1])

    return log_norm

def crf_forward(inputs, state, transition_params, sequence_lengths):
    sequence_lengths = tf.cast(sequence_lengths, dtype=tf.int32)

    last_index = tf.maximum(
        tf.constant(0, dtype=sequence_lengths.dtype), sequence_lengths - 1)
    inputs = tf.transpose(inputs, [1, 0, 2])
    transition_params = tf.expand_dims(transition_params, 0)

    def _scan_fn(_state, _inputs):#递归函数
        _state = tf.expand_dims(_state, 2)
        transition_scores = _state + transition_params# 初始状态往任意点转移的概率
        new_alphas = _inputs + tf.reduce_logsumexp(transition_scores, [1])
        return new_alphas

    all_alphas = tf.transpose(tf.scan(_scan_fn, inputs, state), [1, 0, 2])
    # add first state for sequences of length 1
    all_alphas = tf.concat([tf.expand_dims(state, 1), all_alphas], 1)

    idxs = tf.stack([tf.range(tf.shape(last_index)[0]), last_index], axis=1)
    return tf.gather_nd(all_alphas, idxs)



#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:kaiyinzhou
@Tensorflow 2.0
All of the following Code was follow Google BERT!
"""

from __future__ import absolute_import, division, print_function
import math
import tensorflow as tf
from fennlp.tools import create_initializer, reshape_to_matrix


def transpose_for_scores(input_tensor, batch_size, num_attention_heads, seq_length, width):
    output_tensor = tf.reshape(input_tensor, [batch_size, seq_length, num_attention_heads, width])
    output_tensor = tf.transpose(output_tensor, [0, 2, 1, 3])
    return output_tensor


class MultiAttentionLayer(tf.keras.layers.Layer):
    """
    Performs multi-headed attention from `from_tensor` to `to_tensor`.
    """

    def __init__(self,
                 size_per_head=512,
                 query_act=None,
                 key_act=None,
                 value_act=None,
                 num_attention_heads=1,
                 attention_probs_dropout_prob=0.0,
                 initializer_range=0.02,
                 do_return_2d_tensor=False,
                 batch_size=None,
                 from_seq_length=None,
                 to_seq_length=None,
                 name=None,
                 **kwargs):
        super(MultiAttentionLayer, self).__init__(name=name, **kwargs)
        self.size_per_head = size_per_head
        self.query_act = query_act
        self.key_act = key_act
        self.num_attention_heads = num_attention_heads
        self.value_act = value_act
        self.attention_probs_dropout_prob = attention_probs_dropout_prob
        self.initializer_range = initializer_range
        self.do_return_2d_tensor = do_return_2d_tensor
        self.batch_size = batch_size
        self.from_seq_length = from_seq_length
        self.to_seq_length = to_seq_length

    def build(self, input_shape):
        self.input_spec = tf.keras.layers.InputSpec(shape=input_shape)
        # `query_layer` =[B*F, N*H]
        self._query_layer = tf.keras.layers.Dense(
            self.num_attention_heads * self.size_per_head,
            activation=self.query_act,
            name="query",
            kernel_initializer=create_initializer(self.initializer_range)
        )
        # `value_layer` = [B*T, N*H]
        self._key_layer = tf.keras.layers.Dense(
            self.num_attention_heads * self.size_per_head,
            activation=self.key_act,
            name="key",
            kernel_initializer=create_initializer(self.initializer_range)
        )
        # `query_layer` =[B*T, N*H]
        self._value_layer = tf.keras.layers.Dense(
            self.num_attention_heads * self.size_per_head,
            activation=self.value_act,
            name="value",
            kernel_initializer=create_initializer(self.initializer_range)
        )
        self.drop_out = tf.keras.layers.Dropout(self.attention_probs_dropout_prob)
        self.built = True

    def call(self, from_tensor, to_tensor=None, attention_mask=None, is_training=True):
        from_shape = from_tensor.shape.as_list()
        to_shape = to_tensor.shape.as_list()
        if len(from_shape) != len(to_shape):
            raise ValueError("The rank of `from_tensor` must match the rank of `to_tensor`.")
        if len(from_shape) == 3:
            self.batch_size = from_shape[0]
            self.from_seq_length = from_shape[1]
            self.to_seq_length = to_shape[1]
        elif len(from_shape) == 2:
            if (self.batch_size is None or self.from_seq_length is None or self.to_seq_length is None):
                raise ValueError(
                    "When passing in rank 2 tensors to attention_layer, the values "
                    "for `batch_size`, `from_seq_length`, and `to_seq_length` "
                    "must all be specified."
                )
        from_tensor_2d = reshape_to_matrix(from_tensor)
        to_tensor_2d = reshape_to_matrix(to_tensor)
        query_layer = self._query_layer(from_tensor_2d)
        key_layer = self._key_layer(to_tensor_2d)
        value_layer = self._value_layer(to_tensor_2d)
        # [B,N,F,H]
        query_layer = transpose_for_scores(query_layer, self.batch_size, self.num_attention_heads,
                                           self.from_seq_length, self.size_per_head)
        # [B,N,T,H]
        key_layer = transpose_for_scores(key_layer, self.batch_size, self.num_attention_heads,
                                         self.to_seq_length, self.size_per_head)
        # attention_score = [B,N,F,T]
        attention_score = tf.linalg.matmul(query_layer, key_layer, transpose_b=True)
        attention_score = tf.math.multiply(attention_score, 1.0 / math.sqrt(float(self.size_per_head)))
        if attention_mask is not None:
            # [batch_size, from_seq_length, to_seq_length]
            attention_mask = tf.expand_dims(attention_mask, axis=[1])
            # Here we convert aim position to zero, masked position to -10000
            adder = (1.0 - tf.cast(attention_mask, tf.float32)) * -10000.0
            # This could let score for masked very low
            attention_score += adder
        # sofmax
        attention_probs = tf.math.softmax(attention_score)
        attention_probs = self.drop_out(attention_probs, training=is_training)
        # value [B,N,T,H]
        value_layer = transpose_for_scores(value_layer, self.batch_size, self.num_attention_heads,
                                           self.to_seq_length, self.size_per_head)
        # context_layer [B,N,F,H]
        context_layer = tf.linalg.matmul(attention_probs, value_layer)
        # [B,F,N,H]
        context_layer = tf.transpose(context_layer, [0, 2, 1, 3])
        if self.do_return_2d_tensor:
            context_layer = tf.reshape(context_layer, [self.batch_size * self.from_seq_length,
                                                       self.num_attention_heads * self.size_per_head])
        else:
            context_layer = tf.reshape(context_layer, [self.batch_size, self.from_seq_length,
                                                       self.num_attention_heads * self.size_per_head])
        return context_layer


# Bahdanau2015
class BahdanauAttentionLayer(tf.keras.layers.Layer):
    def __init__(self, hidden_size,
                 v_initializer=None,
                 w_initializer=None,
                 u_initializer=None,
                 use_bias=True,
                 b_initializer='zero',
                 name='att', **kwargs):
        super(BahdanauAttentionLayer, self).__init__(name=name, **kwargs)
        self.hidden_size = hidden_size
        self.use_bias = use_bias
        self.v_initializer = tf.keras.initializers.get(v_initializer)
        self.w_initializer = tf.keras.initializers.get(w_initializer)
        self.u_initializer = tf.keras.initializers.get(u_initializer)
        self.b_initializer = tf.keras.initializers.get(b_initializer)

    def build(self, input_shape):
        inner_size = input_shape.as_list()
        self.V = self.add_weight(
            name="V",
            shape=[self.hidden_size],
            initializer=self.v_initializer,
        )
        self.W = self.add_weight(
            name="W",
            shape=[inner_size[-1], self.hidden_size],
            initializer=self.w_initializer,
        )
        self.U = self.add_weight(
            name="U",
            shape=[inner_size[-1], self.hidden_size],
            initializer=self.u_initializer,
        )
        if self.use_bias:
            self.bias = self.add_weight(
                name="bias",
                shape=[self.hidden_size],
                initializer=self.b_initializer,
            )

        self.attn = tf.keras.layers.Dense(self.hidden_size, activation="tanh")

    def call(self, hidden_state, encoder_outputs,training=False):
        """
        :param hidden_state: [B,D]
        :param encoder_outputs: [B,T,D]
        :return:
        """
        hidden_inner_dim = hidden_state.get_shape().as_list()[-1]
        encoder_inner_dim = encoder_outputs.get_shape().as_list()[-1]
        if hidden_inner_dim != encoder_inner_dim:
            raise ValueError("The last shape of hidden_state and encoder_outputs must equal!")
        dens1 = tf.tensordot(tf.expand_dims(hidden_state, 1), self.W, axes=(2, 1))
        dens2 = tf.tensordot(encoder_outputs, self.U, axes=(2, 1))
        tanh_ = self.attn(dens1 + dens2 + self.bias) if self.use_bias else self.attn(dens1 + dens2)
        cij = tf.tensordot(tanh_, self.V, axes=(2, 0))
        alphas = tf.keras.backend.softmax(cij)
        output = tf.math.reduce_sum(encoder_outputs * tf.expand_dims(alphas, -1), axis=1)  # [B*T*2H]
        return output, alphas

    def get_config(self):
        config = {
            'hidden_size': self.hidden_size,
            "use_bias": self.use_bias,
            'v_initializer': tf.keras.initializers.serialize(self.v_initializer),
            'w_initializer': tf.keras.initializers.serialize(self.w_initializer),
            'u_initializer': tf.keras.initializers.serialize(self.u_initializer),
            'b_initializer': tf.keras.initializers.serialize(self.b_initializer),
        }
        base_config = super(BahdanauAttentionLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape




#! encoding="utf-8"
import tensorflow as tf


class GraphConvolution(tf.keras.layers.Layer):
    def __init__(self,
                 out_features,
                 kernel_initializer='glorot_uniform',
                 kernel_regularizer=None,
                 kernel_constraint=None,
                 bias_initializer='zeros',
                 bias_regularizer=None,
                 bias_constraint=None,
                 use_bias=True,
                 **kwargs):

        super(GraphConvolution, self).__init__(**kwargs)
        self.use_bias = use_bias
        self.out_features = out_features
        self.kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self.kernel_regularizer = tf.keras.regularizers.get(kernel_regularizer)
        self.kernel_constraint = tf.keras.constraints.get(kernel_constraint)
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)
        self.bias_regularizer = tf.keras.regularizers.get(bias_regularizer)
        self.bias_constraint = tf.keras.constraints.get(bias_constraint)

    def build(self, input_shape):
        in_features = input_shape[1]
        self.weight = self.add_weight(
            shape=(in_features, self.out_features),
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            name='kernel',
        )
        if self.use_bias:
            self.bias = self.add_weight(
                shape=(self.out_features,),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                name='bias',
                constraint=self.bias_constraint
            )
        else:
            self.bias = None

        self.build = True

    def call(self, input, adj):
        # input:csr_matrix
        support = tf.linalg.matmul(input, self.weight)
        # SparseTensor format expected by sparse_tensor_dense_matmul: sp_a (indices, values)
        output = tf.sparse.sparse_dense_matmul(adj, support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def get_config(self):
        config = {'units': self.out_features,
                  'use_bias': self.use_bias,

                  'activation': tf.keras.activations.serialize(self.activation),

                  'kernel_initializer': tf.keras.initializers.serialize(
                      self.kernel_initializer),
                  'bias_initializer': tf.keras.initializers.serialize(
                      self.bias_initializer),
                  'kernel_regularizer': tf.keras.regularizers.serialize(
                      self.kernel_regularizer),
                  'bias_regularizer': tf.keras.regularizers.serialize(
                      self.bias_regularizer),
                  'kernel_constraint': tf.keras.constraints.serialize(
                      self.kernel_constraint),
                  'bias_constraint': tf.keras.constraints.serialize(
                      self.bias_constraint)
                  }

        base_config = super(GraphConvolution, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:zhoukaiyin
"""
import re
import time
import tensorflow as tf
import collections
import numpy as np

version = tf.__version__
Version_float = float('.'.join(version.split('.')[:2]))

#For BERT
def create_initializer(initializer_range=0.02):
    """Creates a `truncated_normal_initializer` with the given range."""
    # return tf.keras.initializers.get("truncated_normal")
    return tf.keras.initializers.TruncatedNormal(initializer_range)


def reshape_to_matrix(tensor):
    if len(tensor.shape) == 0:
        return tensor
    dim = tensor.shape[-1]
    tensor_2d = tf.reshape(tensor, [-1, dim])
    return tensor_2d


def reshape_from_matrix(output_tensor, orig_shape_list):
    if len(orig_shape_list) == 2:
        return output_tensor
    output_shape = get_shape_list(output_tensor)
    orig_dims = orig_shape_list[0:-1]
    width = output_shape[-1]
    return tf.reshape(output_tensor, orig_dims + [width])


def create_attention_mask_from_input_mask(from_tensor, to_mask):
    """
    Create 3D attention mask from a 2D tensor mask.
    from_tensor [B,F,D]
    to_mask [B,T]
    """
    assert_rank(from_tensor, [2, 3])
    from_shape = get_shape_list(from_tensor)
    batch_size = from_shape[0]
    from_seq_length = from_shape[1]
    to_shape = get_shape_list(to_mask)
    to_seq_length = to_shape[1]

    to_mask = tf.cast(
        tf.reshape(to_mask, [batch_size, 1, to_seq_length]), tf.float32
    )

    broadcast_ones = tf.ones(
        shape=[batch_size, from_seq_length, 1], dtype=tf.float32
    )

    mask = broadcast_ones * to_mask
    return mask


def assert_rank(tensor, expected_rank, name=None):
    # if name is None:
    #     name = tensor.name
    excepted_rank_dict = {}
    if isinstance(expected_rank, int):
        excepted_rank_dict[expected_rank] = True
    else:
        for x in expected_rank:
            excepted_rank_dict[x] = True
    actual_rank = tensor.shape.ndims
    if actual_rank not in excepted_rank_dict:
        raise (
            "For tensor {} , the actual rank {} is not equal"
            "to expected rank {}".format(name, actual_rank, str(expected_rank))
        )

def gelu(x):
    import numpy as np
    cdf = 0.5 * (1.0 + tf.tanh((np.sqrt(2 / np.pi) * (x + 0.044715 * tf.pow(x, 3)))))
    return x * cdf

def get_activation(activation_string):
    "map string to activation function"
    if not isinstance(activation_string, str):
        return activation_string
    if not activation_string:
        return None
    act = activation_string.lower()
    if act == "linear":
        return None
    elif act == "gelu":
        return gelu
    elif act == "relu":
        return tf.nn.relu
    elif act == "tanh":
        return tf.nn.tanh
    else:
        raise ValueError("Unsupport activation:%s" % act)


def get_assignment_map_from_checkpoint(tvars, init_checkpoint):
    initialized_variable_names = {}
    name_to_variable = collections.OrderedDict()
    for var in tvars:
        print(var.name)
        name = var.name
        m = re.match("^(.*):\\d+$", name)
        if m is not None:
            name = m.group(1)
        name_to_variable[name] = var
    init_vars = tf.train.list_variables(init_checkpoint)
    assignment_map = collections.OrderedDict()
    for x in init_vars:
        (name, var) = (x[0], x[1])
        if name not in name_to_variable:
            continue
        assignment_map[name] = name
        initialized_variable_names[name] = 1
        initialized_variable_names[name + ":0"] = 1
    return (assignment_map, initialized_variable_names)


def init_weights_from_checkpoint(model, checkpoint_file, num_layer,pooler):
    def _loader(checkpoint_file):
        def _loader(name):
            return tf.train.load_variable(checkpoint_file, name)

        return _loader

    loader = _loader(checkpoint_file)
    weights = [
        loader("bert/embeddings/word_embeddings"),
        loader("bert/embeddings/token_type_embeddings"),
        loader("bert/embeddings/position_embeddings"),
        loader("bert/embeddings/LayerNorm/gamma"),
        loader("bert/embeddings/LayerNorm/beta"),
    ]
    encoder = []
    for i in range(num_layer):
        encoder.extend([
            loader("bert/encoder/layer_{}/attention/self/query/kernel".format(i)),
            loader("bert/encoder/layer_{}/attention/self/query/bias".format(i)),
            loader("bert/encoder/layer_{}/attention/self/key/kernel".format(i)),
            loader("bert/encoder/layer_{}/attention/self/key/bias".format(i)),
            loader("bert/encoder/layer_{}/attention/self/value/kernel".format(i)),
            loader("bert/encoder/layer_{}/attention/self/value/bias".format(i)),
            loader("bert/encoder/layer_{}/attention/output/dense/kernel".format(i)),
            loader("bert/encoder/layer_{}/attention/output/dense/bias".format(i)),
            loader("bert/encoder/layer_{}/intermediate/dense/kernel".format(i)),
            loader("bert/encoder/layer_{}/intermediate/dense/bias".format(i)),
            loader("bert/encoder/layer_{}/output/dense/kernel".format(i)),
            loader("bert/encoder/layer_{}/output/dense/bias".format(i)),
            loader("bert/encoder/layer_{}/output/LayerNorm/gamma".format(i)),
            loader("bert/encoder/layer_{}/output/LayerNorm/beta".format(i)),
            loader("bert/encoder/layer_{}/attention/output/LayerNorm/gamma".format(i)),
            loader("bert/encoder/layer_{}/attention/output/LayerNorm/beta".format(i)),
        ])

    weights.extend(encoder)
    if pooler:
        weights.extend([loader("bert/pooler/dense/kernel"),
                    loader("bert/pooler/dense/bias")])

    variables = model.get_layer("bert").variables
    varname = [var.name for var in variables]
    # weight = {name for name in weights}
    for i, var in enumerate(varname):
        print("INIT WEIGHTS {}".format(var), i)

    model.get_layer("bert").set_weights(weights)
    del weights


# calculate function time
def timeit(func):
    def wrapper(*args, **kwargs):
        """
        :param args: 
        :param kwargs: 
        :return:
        """
        start = time.clock()
        result = func(*args, **kwargs)
        end = time.clock()
        print("该函数用时{}".format(end - start))
        return result

    return wrapper


def get_shape_list(tensor, expected_rank=None, name=None):
    # if name is None:
    #     name = tensor.name
    if expected_rank is not None:
        assert_rank(tensor, expected_rank, name)

    shape = tensor.shape.as_list()

    non_static_indexes = []
    for (index, dim) in enumerate(shape):
        if dim is None:
            non_static_indexes.append(index)

    if not non_static_indexes:

        return shape

    dyn_shape = tf.shape(tensor)
    for index in non_static_indexes:
        shape[index] = dyn_shape[index]
    return shape

# For GNN
def uniform(shape,scale=0.05,name=None):
    initial = tf.keras.backend.random_uniform(shape,minval=-scale,maxval=scale,dtype=tf.float32)
    return tf.Variable(initial,name=name)

def glort(shape,name=None):
    initial_range = np.sqrt(6.0/(shape[0]+shape[1]))
    initial = tf.keras.backend.random_uniform(shape,minval=-initial_range,
                                              maxval=initial_range,dtype=tf.float32)
    return tf.Variable(initial,name=name)

def zeros(shape,name=None):
    initial = tf.zeros(shape,dtype=tf.float32)
    return tf.Variable(initial,name=name)

def ones(shape,name=None):
    initial = tf.ones(shape,dtype=tf.float32)
    return tf.Variable(initial,name=name)




import tensorflow as tf


class MaskSparseCategoricalCrossentropy():
    def __init__(self, from_logits):
        self.from_logits = from_logits

    def __call__(self, y_true, y_predict, use_mask=False):
        cross_entropy = tf.keras.losses.sparse_categorical_crossentropy(y_true, y_predict,self.from_logits)

        if use_mask:
            input_mask = tf.cast(tf.math.not_equal(y_true, 0), tf.float32)
            cross_entropy = tf.reduce_sum(cross_entropy * input_mask) / tf.reduce_sum(input_mask)  # mask loss
            return cross_entropy
        else:
            return cross_entropy




#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:zhoukaiyin
"""
import tensorflow as tf
from fennlp.tools import get_activation, get_shape_list, create_attention_mask_from_input_mask
from fennlp.layers.embedding import WDEmbedding, SegPosEmbedding
from fennlp.layers.transformer import Transformer


class BERT(tf.keras.layers.Layer):
    def __init__(self,
                 param=None,
                 use_crf=False,
                 batch_size=2,
                 maxlen=128,
                 vocab_size=21128,
                 hidden_size=768,
                 hidden_act="gelu",
                 num_attention_heads=12,
                 initializer_range=0.02,
                 hidden_dropout_prob=0.1,
                 type_vocab_size=2,
                 intermediate_size=3072,
                 max_position_embeddings=512,
                 attention_probs_dropout_prob=0.1,
                 use_one_hot_embeddings=True,
                 do_return_all_layers=True,
                 num_hidden_layers=12,
                 name=None,
                 **kwargs):
        super(BERT, self).__init__(name=name, **kwargs)
        self.maxlen = param.get("maxlen")
        self.intermediate_size = param.get("intermediate_size")
        self.vocab_size = param.get("vocab_size")
        self.batch_size = param.get("batch_size")
        self.hidden_size = param.get("hidden_size")
        self.hidden_act = param.get("hidden_act")
        self.initializer_range = param.get("initializer_range")
        self.hidden_dropout_prob = param.get("hidden_dropout_prob")
        self.type_vocab_size = param.get("type_vocab_size")
        self.num_attention_heads = param.get("num_attention_heads")
        self.max_position_embeddings = param.get("max_position_embeddings")
        self.attention_probs_dropout_prob = param.get("attention_probs_dropout_prob")
        self.num_hidden_layers = param.get("num_hidden_layers")
        self.use_crf=use_crf

        self.use_one_hot_embeddings = use_one_hot_embeddings
        self.do_return_all_layers = do_return_all_layers

    def build(self, input_shape):
        self.token_embedding = WDEmbedding(vocab_size=self.vocab_size,
                                           embedding_size=self.hidden_size,
                                           initializer_range=self.initializer_range,
                                           word_embedding_name="word_embeddings",
                                           use_one_hot_embedding=self.use_one_hot_embeddings,
                                           name="embeddings")
        # segment and position embedding
        self.segposembedding = SegPosEmbedding(use_token_type=True,
                                               hidden_dropout_prob=self.hidden_dropout_prob,
                                               token_type_vocab_size=self.type_vocab_size,
                                               token_type_embedding_name="token_type_embeddings",
                                               use_position_embeddings=True,
                                               position_embedding_name="position_embeddings",
                                               initializer_range=self.initializer_range,
                                               max_position_embeddings=self.max_position_embeddings,
                                               name="embeddings"
                                               )
        self.encoder_layers = []
        for layer_idx in range(self.num_hidden_layers):
            self.encoder_layer = Transformer(
                batch_size=self.batch_size,
                seq_length=self.maxlen,
                hidden_size=self.hidden_size,
                num_attention_heads=self.num_attention_heads,
                intermediate_size=self.intermediate_size,
                intermediate_act_fn=get_activation(self.hidden_act),
                hidden_dropout_prob=self.attention_probs_dropout_prob,
                initializer_range=self.initializer_range,
                name="layer_{}".format(layer_idx)
            )
            self.encoder_layers.append(self.encoder_layer)

        self.pool_out = tf.keras.layers.Dense(
            self.hidden_size,
            activation=tf.tanh,
            # kernel_constraint=create_initializer(config.initializer_range),
            name="dense"
        )

        self.built = True

    def call(self, inputs, is_training=True):
        input_ids, token_type_ids , input_mask = tf.split(inputs,3,0)
        input_ids = tf.cast(tf.squeeze(input_ids,axis=0), tf.int64)
        token_type_ids = tf.cast(tf.squeeze(token_type_ids,axis=0), tf.int64)
        input_mask = tf.cast(tf.squeeze(input_mask,axis=0), tf.int64)
        input_shape = get_shape_list(input_ids)
        batch_size = input_shape[0]
        seq_length = input_shape[1]
        if input_mask is None:
            input_mask = tf.ones(shape=[batch_size, seq_length], dtype=tf.int32)
        if token_type_ids is None:
            token_type_ids = tf.zeros(shape=[batch_size, seq_length], dtype=tf.int32)
        self.embedding_output = self.token_embedding(input_ids)
        self.embedding_output = self.segposembedding(self.embedding_output, token_type_ids, is_training)
        with tf.keras.backend.name_scope("encoder"):
            attention_mask = create_attention_mask_from_input_mask(input_ids, input_mask)
            all_layer_outputs = []
            layer_encode_output = self.embedding_output
            # print(layer_encode_output)#[3,5,768]
            for encoder_layer in self.encoder_layers:
                layer_encode_input = layer_encode_output
                layer_encode_output = encoder_layer(layer_encode_input, attention_mask, is_training)
                all_layer_outputs.append(layer_encode_output)
            if self.do_return_all_layers:
                self.sequence_output = all_layer_outputs
            else:
                self.sequence_output = layer_encode_output
        self.sequence_output = self.sequence_output[-1]

        return self

    def get_pooled_output(self):
        with tf.keras.backend.name_scope("pooler"):
            first_token_tensor = tf.squeeze(self.sequence_output[:, 0:1, :], axis=1)
            self.pooled_output = self.pool_out(first_token_tensor)
        return self.pooled_output

    def get_sequence_output(self):
        return self.sequence_output

    def get_all_encoder_layers(self):
        return self.all_encoder_layers

    def get_embedding_output(self):
        return self.embedding_output



#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:zhoukaiyin
"""
import tensorflow as tf


class BiLSTM(tf.keras.layers.Layer):
    def __init__(self, hidden_dim, vocab_size, embedding_size, dropout_rate, maxlen, **kwargs):
        super(BiLSTM, self).__init__(**kwargs)
        self.embed = tf.keras.layers.Embedding(vocab_size + 1, embedding_size, maxlen, mask_zero=True)
        self.bilstm = tf.keras.layers.Bidirectional(
            tf.keras.layers.LSTM(hidden_dim,
                                 dropout=dropout_rate)
        )

    def call(self, inputs):
        embed = self.embed(inputs)
        logits = self.bilstm(embed)
        return logits



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implements F scores."""

import tensorflow as tf
from typeguard import typechecked
from sklearn.metrics import recall_score, precision_score, f1_score, accuracy_score
from collections import defaultdict
from .type import AcceptableDTypes, FloatTensorLike
from typing import Optional

import warnings
import numpy as np

warnings.filterwarnings('ignore')


class FBetaScore(tf.keras.metrics.Metric):
    """Computes F-Beta score.
    It is the weighted harmonic mean of precision
    and recall. Output range is [0, 1]. Works for
    both multi-class and multi-label classification.
    F-Beta = (1 + beta^2) * (prec * recall) / ((beta^2 * prec) + recall)
    Args:
        num_classes: Number of unique classes in the dataset.
        average: Type of averaging to be performed on data.
            Acceptable values are `None`, `micro`, `macro` and
            `weighted`. Default value is None.
        beta: Determines the weight of precision and recall
            in harmonic mean. Determines the weight given to the
            precision and recall. Default value is 1.
        threshold: Elements of `y_pred` greater than threshold are
            converted to be 1, and the rest 0. If threshold is
            None, the argmax is converted to 1, and the rest 0.
    Returns:
        F-Beta Score: float
    Raises:
        ValueError: If the `average` has values other than
        [None, micro, macro, weighted].
        ValueError: If the `beta` value is less than or equal
        to 0.
    `average` parameter behavior:
        None: Scores for each class are returned
        micro: True positivies, false positives and
            false negatives are computed globally.
        macro: True positivies, false positives and
            false negatives are computed for each class
            and their unweighted mean is returned.
        weighted: Metrics are computed for each class
            and returns the mean weighted by the
            number of true instances in each class.
    """

    @typechecked
    def __init__(
            self,
            num_classes: FloatTensorLike,
            average: Optional[str] = None,
            beta: FloatTensorLike = 1.0,
            threshold: Optional[FloatTensorLike] = None,
            name: str = "fbeta_score",
            dtype: AcceptableDTypes = None,
            **kwargs
    ):
        super().__init__(name=name, dtype=dtype)

        if average not in (None, "micro", "macro", "weighted"):
            raise ValueError(
                "Unknown average type. Acceptable values "
                "are: [None, micro, macro, weighted]"
            )

        if not isinstance(beta, float):
            raise TypeError("The value of beta should be a python float")

        if beta <= 0.0:
            raise ValueError("beta value should be greater than zero")

        if threshold is not None:
            if not isinstance(threshold, float):
                raise TypeError("The value of threshold should be a python float")
            if threshold > 1.0 or threshold <= 0.0:
                raise ValueError("threshold should be between 0 and 1")

        self.num_classes = num_classes
        self.average = average
        self.beta = beta
        self.threshold = threshold
        self.axis = None
        self.init_shape = []

        if self.average != "micro":
            self.axis = 0
            self.init_shape = [self.num_classes]

        def _zero_wt_init(name):
            return self.add_weight(
                name, shape=self.init_shape, initializer="zeros", dtype=self.dtype
            )

        self.true_positives = _zero_wt_init("true_positives")
        self.false_positives = _zero_wt_init("false_positives")
        self.false_negatives = _zero_wt_init("false_negatives")
        self.weights_intermediate = _zero_wt_init("weights_intermediate")

    # TODO: Add sample_weight support, currently it is
    # ignored during calculations.
    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true = tf.reshape(y_true, shape=[-1])
        y_true = tf.one_hot(y_true, self.num_classes)
        y_pred = tf.reshape(y_pred, shape=[-1, self.num_classes])

        if self.threshold is None:
            threshold = tf.reduce_max(y_pred, axis=-1, keepdims=True)
            # make sure [0, 0, 0] doesn't become [1, 1, 1]
            # Use abs(x) > eps, instead of x != 0 to check for zero
            y_pred = tf.logical_and(y_pred >= threshold, tf.abs(y_pred) > 1e-12)
        else:
            y_pred = y_pred > self.threshold

        y_true = tf.cast(y_true, tf.int32)
        y_pred = tf.cast(y_pred, tf.int32)

        def _count_non_zero(val):
            non_zeros = tf.math.count_nonzero(val, axis=self.axis)
            return tf.cast(non_zeros, self.dtype)

        self.true_positives.assign_add(_count_non_zero(y_pred * y_true))
        self.false_positives.assign_add(_count_non_zero(y_pred * (y_true - 1)))
        self.false_negatives.assign_add(_count_non_zero((y_pred - 1) * y_true))
        self.weights_intermediate.assign_add(_count_non_zero(y_true))

    def result(self):
        precision = tf.math.divide_no_nan(
            self.true_positives, self.true_positives + self.false_positives
        )
        recall = tf.math.divide_no_nan(
            self.true_positives, self.true_positives + self.false_negatives
        )

        mul_value = precision * recall
        add_value = (tf.math.square(self.beta) * precision) + recall
        mean = tf.math.divide_no_nan(mul_value, add_value)
        f1_score = mean * (1 + tf.math.square(self.beta))

        if self.average == "weighted":
            weights = tf.math.divide_no_nan(
                self.weights_intermediate, tf.reduce_sum(self.weights_intermediate)
            )
            f1_score = tf.reduce_sum(f1_score * weights)

        elif self.average is not None:  # [micro, macro]
            f1_score = tf.reduce_mean(f1_score)

        return f1_score

    def get_config(self):
        """Returns the serializable config of the metric."""

        config = {
            "num_classes": self.num_classes,
            "average": self.average,
            "beta": self.beta,
        }

        if self.threshold is not None:
            config["threshold"] = self.threshold

        base_config = super().get_config()
        return {**base_config, **config}

    def reset_states(self):
        self.true_positives.assign(tf.zeros(self.init_shape, self.dtype))
        self.false_positives.assign(tf.zeros(self.init_shape, self.dtype))
        self.false_negatives.assign(tf.zeros(self.init_shape, self.dtype))
        self.weights_intermediate.assign(tf.zeros(self.init_shape, self.dtype))


class KSparseF1Score(FBetaScore):
    """Computes F-1 Score.
    It is the harmonic mean of precision and recall.
    Output range is [0, 1]. Works for both multi-class
    and multi-label classification.
    F-1 = 2 * (precision * recall) / (precision + recall)
    Args:
        num_classes: Number of unique classes in the dataset.
        average: Type of averaging to be performed on data.
            Acceptable values are `None`, `micro`, `macro`
            and `weighted`. Default value is None.
        threshold: Elements of `y_pred` above threshold are
            considered to be 1, and the rest 0. If threshold is
            None, the argmax is converted to 1, and the rest 0.
    Returns:
        F-1 Score: float
    Raises:
        ValueError: If the `average` has values other than
        [None, micro, macro, weighted].
    `average` parameter behavior:
        None: Scores for each class are returned
        micro: True positivies, false positives and
            false negatives are computed globally.
        macro: True positivies, false positives and
            false negatives are computed for each class
            and their unweighted mean is returned.
        weighted: Metrics are computed for each class
            and returns the mean weighted by the
            number of true instances in each class.
    """

    @typechecked
    def __init__(
            self,
            num_classes: FloatTensorLike,
            average: str = None,
            threshold: Optional[FloatTensorLike] = None,
            name: str = "f1_score",
            dtype: AcceptableDTypes = None,
            **kwargs
    ):
        super().__init__(num_classes, average, 1.0, threshold, name=name, dtype=dtype)

    def get_config(self):
        base_config = super().get_config()
        del base_config["beta"]
        return base_config


class SparseF1Score(object):
    def __init__(self, average, predict_sparse=False):
        self.average = average
        self.predict_sparse = predict_sparse

    def __call__(self, y_true, y_predict):
        y_true = tf.reshape(tf.constant(y_true), [-1]).numpy()

        if self.predict_sparse:
            y_predict = tf.reshape(y_predict, [-1]).numpy()
        else:
            y_predict = tf.reshape(tf.argmax(y_predict, -1), [-1]).numpy()

        f1 = f1_score(y_true, y_predict, average=self.average)
        return f1


class SparseRecallScore(object):
    def __init__(self, average, predict_sparse=False):
        self.average = average
        self.predict_sparse = predict_sparse

    def __call__(self, y_true, y_predict):
        if self.predict_sparse:
            y_predict = tf.reshape(y_predict, [-1]).numpy()
        else:
            y_predict = tf.reshape(tf.argmax(y_predict, -1), [-1]).numpy()
        y_true = tf.reshape(tf.constant(y_true), [-1]).numpy()
        recall = recall_score(y_true, y_predict, average=self.average)
        return recall


class SparsePrecisionScore(object):
    def __init__(self, average, predict_sparse=False):
        self.average = average
        self.predict_sparse = predict_sparse

    def __call__(self, y_true, y_predict):
        y_true = tf.reshape(tf.constant(y_true), [-1]).numpy()
        if self.predict_sparse:
            y_predict = tf.reshape(y_predict, [-1]).numpy()

        else:
            y_predict = tf.reshape(tf.argmax(y_predict, -1), [-1]).numpy()
        precision = precision_score(y_true, y_predict, average=self.average)
        return precision


class SparseAccuracy(object):
    def __init__(self, predict_sparse=False):
        self.predict_sparse = predict_sparse

    def __call__(self, y_true, y_predict):
        y_true = tf.reshape(tf.constant(y_true), [-1]).numpy()
        if self.predict_sparse:
            y_predict = tf.reshape(y_predict, [-1]).numpy()

        else:
            y_predict = tf.reshape(tf.argmax(y_predict, -1), [-1]).numpy()
        acc = accuracy_score(y_true, y_predict)
        return acc


class HitN_MR_MRR():
    def __init__(self, loader, mode="valid"):
        self.mode = mode
        self.loader = loader
        if self.mode == "train":
            self.data_idxs = loader.get_data_idxs(loader.train_data)
        elif self.mode == "valid":
            self.data_idxs = loader.get_data_idxs(loader.valid_data)  # valid数据集合中的三元组
        elif self.mode == "test":
            self.data_idxs = loader.get_data_idxs(loader.test_data)
        # er_vocb={(h,r):[t1,t2]},train+valid
        self.er_vocab = loader.get_er_vocab(loader.get_data_idxs(loader.data))


    def __call__(self, model, batch_size):
        ranks = []
        hits = defaultdict(list)
        for i in range(10):
            hits[i] = []
        for data in self.loader.get_batch(self.er_vocab, self.data_idxs, batch_size=batch_size):
            # actually here h_r is h_r_t
            h_idx = data['h_r'][:, 0]
            r_idx = data['h_r'][:, 1]
            t_idx = data['h_r'][:, 2]
            hrt = data["h_r"].numpy()
            predict = model.predict(h_idx, r_idx)
            predict = predict.numpy()
            for j in range(hrt.shape[0]):
                # 也就是说对于当前样本仅仅考虑目标尾实体的得分，而其他的也是真实尾实体得分被置为0。
                filt = self.er_vocab[(hrt[j][0], hrt[j][1])]
                target_value = predict[j, t_idx[j]]
                predict[j, filt] = 0.0
                predict[j, t_idx[j]] = target_value
            sort_idxs = tf.argsort(predict, axis=1, direction='DESCENDING')
            del predict
            for j in range(hrt.shape[0]):
                rank = np.where(sort_idxs[j] == t_idx[j])[0][0]
                ranks.append(rank + 1)
                for hits_level in range(10):
                    if rank <= hits_level:
                        hits[hits_level].append(1.0)
                    else:
                        hits[hits_level].append(0.0)
            del sort_idxs

        hit10 = np.mean(hits[9])
        hit5 = np.mean(hits[4])
        hit3 = np.mean(hits[2])
        hit1 = np.mean(hits[0])
        MR = np.mean(ranks)
        MRR =np.mean(1./np.array(ranks))
        return hit1, hit3, hit5, hit10, MR, MRR



#! encoding="utf-8"
import tensorflow as tf
from fennlp.gnn import GCNConv


class GCN2Layer(tf.keras.Model):
    def __init__(self, hidden_dim, num_class, dropout_rate=0.5, **kwargs):
        super(GCN2Layer, self).__init__(**kwargs)
        self.hidden_dim = hidden_dim
        self.num_class = num_class
        self.gc1 = GCNConv.GraphConvolution(hidden_dim)
        self.gc2 = GCNConv.GraphConvolution(num_class)
        self.dropout = tf.keras.layers.Dropout(dropout_rate)

    def call(self, inputs, adj, training=True):
        x = tf.nn.relu(self.gc1(inputs, adj))
        x = self.dropout(x, training=training)
        x = self.gc2(x, adj)
        return tf.math.softmax(x, -1)

    def predict(self, inputs, adj, training=False):
        return self(inputs, adj, training)



import numpy as np
import tensorflow as tf
import os
from fennlp.tokenizers import tokenization
import collections
import codecs
import pickle


class ZHTFWriter(object):
    def __init__(self, maxlen, vocab_files, modes, task="NER", do_low_case=True,
                 check_exist=False):
        self.maxlen = maxlen
        self.fulltoknizer = tokenization.FullTokenizer(
            vocab_file=vocab_files, do_lower_case=do_low_case
        )
        for mode in modes:
            self.mode = mode
            print("Writing {}".format(self.mode))
            self.filename = os.path.join("Input", self.mode)
            if task == "NER":
                self.write(mode, check_exist, task)
            elif task == "cls":
                self.write(mode, check_exist, task)

    def write(self, mode, check_exist, task):
        if check_exist:
            if os.path.exists(self.filename + ".tfrecords"):
                self.label_map = pickle.load(open(os.path.join("Input", "label2id.pkl"), 'rb'))
                print("Having Writen {} file in to device successfully!".format(mode))
                pass
            else:
                examples = self._read_file()
                self.label_map = self.label2id()
                self._write_examples(examples, task)
        else:
            examples = self._read_file()
            self.label_map = self.label2id()
            self._write_examples(examples, task)

    def _read_file(self):
        with codecs.open(self.filename, encoding='utf-8') as rf:
            examples = self._creat_examples(rf, self.mode)
        return examples

    def _creat_examples(self, lines, mode):
        examples = []
        self.label_list = set()
        for line in lines:
            line = line.strip().split('\t')
            if mode == "test":
                w = tokenization.convert_to_unicode(line[0])
                label = "0"
            else:
                w = tokenization.convert_to_unicode(line[0])
                label = tokenization.convert_to_unicode(line[-1])
            examples.append((w, label))
            self.label_list.update(set(label.split()))
        self.label_list = sorted(self.label_list)
        print("Totally use {} labels!\n".format(len(self.label_list)))
        return examples

    def _creat_features(self, values):
        f = tf.train.Feature(int64_list=tf.train.Int64List(value=list(values)))
        return f

    def label2id(self):
        label_map = {}
        for (i, label) in enumerate(self.label_list):
            label_map[label] = i
        pickle.dump(label_map, open(os.path.join("Input", "label2id.pkl"), 'wb'))
        return label_map

    def _convert_single_example_ner(self, example, maxlen):
        # self.label_map =

        tokens = ["[CLS]"]
        segment_ids = [0]
        input_mask = [1]
        label_id = [self.label_map.get("O")]
        sentences, labels = example
        labels = labels.split()
        sentences = [self.fulltoknizer.tokenize(w) for w in sentences.split()]
        for i, words in enumerate(sentences):
            for word in words:
                if len(tokens) < maxlen - 1:
                    tokens.append(word)
                    segment_ids.append(0)
                    input_mask.append(1)
                    label_id.append(self.label_map.get(labels[i]))
        tokens.append("[SEP]")
        segment_ids.append(0)
        input_mask.append(1)
        label_id.append(self.label_map.get("O"))
        input_ids = self.fulltoknizer.convert_tokens_to_ids(tokens)
        while len(input_ids) < maxlen:
            input_ids.append(0)
            input_mask.append(0)
            segment_ids.append(0)
            label_id.append(self.label_map.get("O"))
        return input_ids, segment_ids, input_mask, label_id

    def _convert_single_example_cls(self, example, maxlen):
        tokens = ["[CLS]"]
        segment_ids = [0]
        input_mask = [1]
        sentences, label = example
        sentences = [self.fulltoknizer.tokenize(w) for w in sentences.split()]
        for i, words in enumerate(sentences):
            for word in words:
                if len(tokens) < maxlen - 1:
                    tokens.append(word)
                    segment_ids.append(0)
                    input_mask.append(1)
        tokens.append("[SEP]")
        segment_ids.append(0)
        input_mask.append(1)
        input_ids = self.fulltoknizer.convert_tokens_to_ids(tokens)
        while len(input_ids) < maxlen:
            input_ids.append(0)
            input_mask.append(0)
            segment_ids.append(0)
        return input_ids, segment_ids, input_mask, [int(label)]


    def _write_examples(self, examples, task="ner"):
        writer = tf.io.TFRecordWriter(os.path.join("Input", self.mode + ".tfrecords"))
        features = collections.OrderedDict()
        for example in examples:
            if task.lower() == "ner":
                input_ids, segment_ids, input_mask, label_id = self._convert_single_example_ner(example, self.maxlen)
            elif task.lower() == 'cls':
                input_ids, segment_ids, input_mask, label_id = self._convert_single_example_cls(example, self.maxlen)
            features["input_ids"] = self._creat_features(input_ids)
            features["label_id"] = self._creat_features(label_id)
            features["segment_ids"] = self._creat_features(segment_ids)
            features["input_mask"] = self._creat_features(input_mask)

            example = tf.train.Example(features=tf.train.Features(feature=features))
            writer.write(example.SerializeToString())
        writer.close()

    def convert_id_to_vocab(self, items):
        vocabs = self.fulltoknizer.convert_ids_to_tokens(items)
        return vocabs

    def convert_id_to_label(self, items):
        id2label = {value: key for key, value in self.label_map.items()}
        output = []
        for item in items:
            output.append(id2label[item])
        return output


class ZHTFLoader(object):
    def __init__(self, maxlen, batch_size, task="ner",epoch=None):
        self.maxlen = maxlen
        self.batch_size = batch_size
        self.epoch = epoch
        self.task=task

    def decode_record(self, record):
        # 告诉解码器每一个feature的类型
        if self.task.lower()=='ner':

            feature_description = {
                "input_ids": tf.io.FixedLenFeature([self.maxlen], tf.int64),
                "label_id": tf.io.FixedLenFeature([self.maxlen], tf.int64),
                "segment_ids": tf.io.FixedLenFeature([self.maxlen], tf.int64),
                "input_mask": tf.io.FixedLenFeature([self.maxlen], tf.int64),

            }

        elif self.task.lower()=="cls":
            feature_description = {
                "input_ids": tf.io.FixedLenFeature([self.maxlen], tf.int64),
                "label_id": tf.io.FixedLenFeature([], tf.int64),
                "segment_ids": tf.io.FixedLenFeature([self.maxlen], tf.int64),
                "input_mask": tf.io.FixedLenFeature([self.maxlen], tf.int64),

            }

        example = tf.io.parse_single_example(record, feature_description)
        return example["input_ids"], example["segment_ids"], example["input_mask"], example["label_id"]

    def load_train(self):
        self.filename = os.path.join("Input", "train.tfrecords")
        raw_dataset = tf.data.TFRecordDataset(self.filename)
        dataset = raw_dataset.apply(
            tf.data.experimental.map_and_batch(
                lambda record: self.decode_record(record),
                batch_size=self.batch_size,
                drop_remainder=True))
        dataset = dataset.apply(
            tf.data.experimental.shuffle_and_repeat(
                1000,
                self.epoch
            )
        )
        return dataset

    def load_valid(self):
        self.filename = os.path.join("Input", "valid.tfrecords")
        raw_dataset = tf.data.TFRecordDataset(self.filename)
        dataset = raw_dataset.apply(
            tf.data.experimental.map_and_batch(
                lambda record: self.decode_record(record),
                batch_size=self.batch_size,
                drop_remainder=True))
        return dataset



#! encoding:utf-8
import tensorflow as tf

constraints = tf.keras.constraints
initializers = tf.keras.initializers
regularizers = tf.keras.regularizers
math_ops = tf.math


class FPNNormalization(tf.keras.layers.Layer):
    """
    article：Filter Response Normalization Layer: Eliminating Batch Dependence in the Training of Deep Neural Networks.
    """

    def __init__(self,
                 epsilon=1e-6,
                 gamma_initializer='ones',
                 beta_initializer='ones',
                 tau_initializer='ones',
                 gamma_regularizer=None,
                 beta_regularizer=None,
                 tau_regularizer=None,
                 gamma_constraint=None,
                 beta_constraint=None,
                 tau_constraint=None,
                 trainable=True,
                 name=None,
                 **kwargs):
        super(FPNNormalization, self).__init__(
            name=name, trainable=trainable, **kwargs)
        self.epsilon = epsilon
        self.gamma_initializer = initializers.get(gamma_initializer)
        self.beta_initializer = initializers.get(beta_initializer)
        self.tau_initializer = initializers.get(tau_initializer)

        self.gamma_regularizer = regularizers.get(gamma_regularizer)
        self.beta_regularizer = regularizers.get(beta_regularizer)
        self.tau_regularizer = regularizers.get(tau_regularizer)

        self.gamma_constraint = constraints.get(gamma_constraint)
        self.beta_constraint = constraints.get(beta_constraint)
        self.tau_constraint = constraints.get(tau_constraint)

    def build(self, input_shape):
        ndims = len(input_shape)
        if ndims is None:
            raise ValueError('Input shape %s has undefined rank.' % input_shape)
        shape = [1] * (ndims - 1)
        shape.append(input_shape[-1])
        self.gamma = self.add_weight(
            shape=shape,
            initializer=self.gamma_initializer,
            regularizer=self.gamma_regularizer,
            constraint=self.gamma_constraint,
            name='gamma',
        )
        self.beta = self.add_weight(
            shape=shape,
            initializer=self.beta_initializer,
            regularizer=self.beta_regularizer,
            constraint=self.beta_constraint,
            name='beta',
        )

        self.tau = self.add_weight(
            shape=shape,
            initializer=self.tau_initializer,
            regularizer=self.tau_regularizer,
            constraint=self.tau_constraint,
            name='tau',
        )
        self.built = True

    def call(self, x):
        nu2 = math_ops.reduce_mean(math_ops.square(x), axis=[1, 2], keepdims=True)
        x = x * math_ops.sqrt(nu2 + math_ops.abs(self.epsilon))
        return math_ops.maximum(self.gamma * x + self.tau, self.gamma)

    def get_config(self):
        config = {
            'epsilon': self.epsilon,
            'gamma_initializer': initializers.serialize(self.gamma_initializer),
            'beta_initializer': initializers.serialize(self.beta_initializer),
            'tau_initializer': initializers.serialize(self.tau_initializer),

            'gamma_regularizer': regularizers.serialize(self.gamma_regularizer),
            'beta_regularizer': regularizers.serialize(self.beta_regularizer),
            'tau_regularizer': regularizers.serialize(self.tau_regularizer),

            'gamma_constraint': constraints.serialize(self.gamma_constraint),
            'beta_constraint': constraints.serialize(self.beta_constraint),
            'tau_constraint': constraints.serialize(self.tau_constraint),
        }
        base_config = super(FPNNormalization, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape

class BatchNormalization(tf.keras.layers.Layer):
    def __init__(self):
        pass


#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
参考：
https://github.com/CyberZHG/keras-bert
https://github.com/tensorflow/tensorflow/blob/v2.1.0/tensorflow/python/keras/backend.py#L1561-L1563
https://github.com/bojone/bert4keras
"""
import tensorflow as tf


class Adam(tf.keras.optimizers.Optimizer):
    def __init__(self,
                 learning_rate=0.01,
                 min_lr=1e-9,
                 beta_1=0.9,
                 beta_2=0.999,
                 epsilon=1e-6,
                 bias_correction=True,
                 name='Adam',
                 **kwargs):
        """
        :param warmup_steps: 学习率在指定的步数线性增长到目标学习率
        :param learning_rate:
        :param beta_1:
        :param beta_2:
        :param epsilon:
        :param bias_correction:
        :param name:
        :param kwargs:
        """
        super(Adam, self).__init__(name, **kwargs)
        self._set_hyper('learning_rate', learning_rate)
        self._set_hyper('min_lr', min_lr)
        self._set_hyper('beta_1', beta_1)
        self._set_hyper('beta_2', beta_2)
        self.epsilon = epsilon or tf.keras.backend.epislon()
        self.bias_correction = bias_correction

    def _create_slots(self, var_list):
        for var in var_list:
            self.add_slot(var, 'm')
            self.add_slot(var, 'v')

    def _resource_apply_op(self, grad, var, indices=None):
        # 准备变量
        var_dtype = var.dtype.base_dtype
        lr_t = self._decayed_lr(var_dtype)
        m = self.get_slot(var, 'm')
        v = self.get_slot(var, 'v')
        beta_1_t = self._get_hyper('beta_1', var_dtype)
        beta_2_t = self._get_hyper('beta_2', var_dtype)
        min_lr = self._get_hyper('min_lr', var_dtype)
        epsilon_t = tf.cast(self.epsilon, var_dtype)
        local_step = tf.cast(self.iterations + 1, var_dtype)
        beta_1_t_power = tf.math.pow(beta_1_t, local_step)
        beta_2_t_power = tf.math.pow(beta_2_t, local_step)

        if indices is None:
            # update 与 state_ops.assign(x,new_x)相同均是赋值含义
            m_t = tf.keras.backend.update(m, beta_1_t * m + (1 - beta_1_t) * grad)
            v_t = tf.keras.backend.update(v, beta_2_t * v + (1 - beta_2_t) * grad ** 2)
        else:
            mv_ops = [tf.keras.backend.update(m, beta_1_t * m), tf.keras.backend.update(v, beta_2_t * v)]
            with tf.control_dependencies(mv_ops):
                m_t = self._resource_scatter_add(m, indices,
                                                 (1 - beta_1_t) * grad)
                v_t = self._resource_scatter_add(v, indices,
                                                 (1 - beta_2_t) * grad ** 2)

        with tf.control_dependencies([m_t, v_t]):
            # 偏差修正
            if self.bias_correction:
                m_t = m_t / (1. - beta_1_t_power)
                v_t = v_t / (1. - beta_2_t_power)
            var_t = var - lr_t * m_t / (tf.keras.backend.sqrt(v_t) + epsilon_t)
            return tf.keras.backend.update(var, var_t)

    def _resource_apply_dense(self, grad, var):
        return self._resource_apply_op(grad, var)

    def _resource_apply_sparse(self, grad, var, indices):
        return self._resource_apply_op(grad, var, indices)

    def get_config(self):
        config = {
            'learning_rate': self._serialize_hyperparameter('learning_rate'),
            'beta_1': self._serialize_hyperparameter('beta_1'),
            'beta_2': self._serialize_hyperparameter('beta_2'),
            'epsilon': self.epsilon,
        }
        base_config = super(Adam, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))




import io
import os
import sys
from setuptools import find_packages, setup

# Package meta-data.
NAME = 'fennlp'
DESCRIPTION = 'An out-of-the-box NLP toolkit can easily help you solve tasks such as\
              Entity Recognition, Text Classification, Relation Extraction and so on.'
URL = 'https://github.com/kyzhouhzau/fennlp'
EMAIL = 'zhoukaiyinhzau@gmail.com'
AUTHOR = 'Kaiyin Zhou'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = '0.0.4'

REQUIRED = [
    'typeguard',
    'gensim',
    'tqdm'
]

# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    install_requires=REQUIRED,
    include_package_data=True,
    license='Apache',
    classifiers=[
        'Programming Language :: Python',
    ],
)


#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:zhoukaiyin
"""
import tensorflow as tf


class TextCNN(tf.keras.Model):
    def __init__(self,
                 maxlen,
                 vocab_size,
                 embedding_dims,
                 class_num=2,
                 kernel_sizes=[2, 3, 4],
                 filter_size=[128, 128, 128],
                 name=None,
                 **kwargs):
        super(TextCNN, self).__init__(name=name, **kwargs)
        self.maxlen = maxlen
        self.kernel_sizes = kernel_sizes
        # "mask_zero: 是否把 0 看作为一个应该被遮蔽的特殊的 "padding" 值。
        # 这对于可变长的 循环神经网络层 十分有用。 如果设定为 True，
        # 那么接下来的所有层都必须支持 masking，否则就会抛出异常。
        # 如果 mask_zero 为 True，作为结果，索引 0 就不能被用于词汇表中
        # （input_dim 应该与 vocabulary + 1 大小相同）。"
        self.embedding = tf.keras.layers.Embedding(vocab_size + 1, embedding_dims,
                                                   input_length=maxlen,
                                                   mask_zero=True)
        self.convs = []
        self.max_poolings = []
        for i, k in enumerate(kernel_sizes):
            # Conv1D inputs=[None,maxlen,embedding_dim]
            self.convs.append(tf.keras.layers.Conv1D(filter_size[i],
                                                     k,
                                                     activation="relu"))
            # here a little different from methods provide in article,
            # but they have the same purpose
            self.max_poolings.append(tf.keras.layers.GlobalAvgPool1D())
        self.dense = tf.keras.layers.Dense(class_num, activation='softmax')

    def call(self, inputs):
        embedding = self.embedding(inputs)
        convs = []
        for i, k in enumerate(self.kernel_sizes):
            out = self.convs[i](embedding)
            out = self.max_poolings[i](out)
            convs.append(out)

        out = tf.keras.layers.concatenate(convs)
        out = self.dense(out)
        return out

    def predict(self, inputs):
        re = self(inputs)
        return re



#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:zhoukaiyin
"""
import tensorflow as tf
from fennlp.tools import create_initializer
from .attention import MultiAttentionLayer
from fennlp.tools import reshape_to_matrix, get_shape_list, reshape_from_matrix


class Transformer(tf.keras.layers.Layer):
    def __init__(self,
                 # layer_idx=0,
                 batch_size,
                 seq_length,
                 hidden_size=768,
                 num_hidden_layers=12,
                 num_attention_heads=12,
                 intermediate_size=3072,
                 intermediate_act_fn="gelu",
                 hidden_dropout_prob=0.1,
                 attention_probs_dropout_prob=0.1,
                 initializer_range=0.02,
                 name=None,
                 **kwargs):
        super(Transformer, self).__init__(name=name, **kwargs)
        self.hidden_size = hidden_size
        self.batch_size = batch_size
        self.seq_length = seq_length
        self.num_hidden_layers = num_hidden_layers
        self.num_attention_heads = num_attention_heads
        self.intermediate_size = intermediate_size
        self.intermediate_act_fn = intermediate_act_fn
        self.hidden_dropout_prob = hidden_dropout_prob
        self.attention_probs_dropout_prob = attention_probs_dropout_prob
        self.initializer_range = initializer_range

    def build(self, input_shape):
        self.input_spec = tf.keras.layers.InputSpec(shape=input_shape)

        attention_head_size = int(self.hidden_size / self.num_attention_heads)
        self.attention_layer = MultiAttentionLayer(
            num_attention_heads=self.num_attention_heads,
            size_per_head=attention_head_size,
            attention_probs_dropout_prob=self.attention_probs_dropout_prob,
            initializer_range=self.initializer_range,
            do_return_2d_tensor=True,
            batch_size=self.batch_size,
            from_seq_length=self.seq_length,
            to_seq_length=self.seq_length,
            name="self"

        )
        self.attention_output_layer = tf.keras.layers.Dense(
            self.hidden_size,
            kernel_initializer=create_initializer(self.initializer_range),
            name="dense"
        )
        self.inter_output = tf.keras.layers.Dense(
            self.intermediate_size,
            activation=self.intermediate_act_fn,
            kernel_initializer=create_initializer(self.initializer_range),
            name="dense"
        )
        self.layer_out = tf.keras.layers.Dense(
            self.hidden_size,
            kernel_initializer=create_initializer(self.initializer_range),
            name="dense"
        )
        self.dropout = tf.keras.layers.Dropout(self.hidden_dropout_prob)
        self.layer_norm = tf.keras.layers.LayerNormalization(axis=-1, name="LayerNorm")
        self.out_layer_norm = tf.keras.layers.LayerNormalization(axis=-1, name="LayerNorm")

        self.built = True

    def call(self, input_tensor, attention_mask=None, is_training=True):
        if self.hidden_size % self.num_attention_heads != 0:
            raise ValueError(
                "The hidden size must be the integer multiple of num attention heads"
            )
        input_shape = get_shape_list(input_tensor)
        input_tensor = reshape_to_matrix(input_tensor)  # [15,768]
        # print(attention_mask)  # [3,5,5]
        # --------------------------------------------------------------------------------------
        with tf.keras.backend.name_scope("attention"):
            attention_heads = []
            attention_head = self.attention_layer(input_tensor, input_tensor, attention_mask)
            attention_heads.append(attention_head)
            if len(attention_heads) == 1:
                attention_output = attention_heads[0]
            else:
                attention_output = tf.concat(attention_heads, axis=-1)
            with tf.keras.backend.name_scope("output"):
                attention_output = self.attention_output_layer(attention_output)
                attention_output = self.dropout(attention_output, training=is_training)
                attention_output = self.out_layer_norm(attention_output + input_tensor)
        with tf.keras.backend.name_scope("intermediate"):
            intermediate_output = self.inter_output(attention_output)
        with tf.keras.backend.name_scope("output"):
            layer_output = self.layer_out(intermediate_output)
            layer_output = self.dropout(layer_output, training=is_training)
            layer_output = self.layer_norm(layer_output + attention_output)
        layer_output = reshape_from_matrix(layer_output, input_shape)
        return layer_output



#! usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@Author:Kaiyin Zhou
"""

import tensorflow as tf

import numpy as np



class TuckER(tf.keras.Model):
    # loader.data
    def __init__(self, loader,
                 ent_vec_dim=200,
                 rel_vec_dim=30,
                 input_dropout=0.2,
                 hidden_dropout1=0.1,
                 hidden_dropout2=0.2,
                 **kwargs):
        super(TuckER, self).__init__(**kwargs)
        self.rel_vec_dim = rel_vec_dim
        self.ent_vec_dim = ent_vec_dim
        self.E = tf.keras.layers.Embedding(len(loader.entities),
                                           ent_vec_dim)
        self.R = tf.keras.layers.Embedding(len(loader.relations),
                                           rel_vec_dim)

        self.input_dropout = tf.keras.layers.Dropout(input_dropout)
        self.hidden_dropout1 = tf.keras.layers.Dropout(hidden_dropout1)
        self.hidden_dropout2 = tf.keras.layers.Dropout(hidden_dropout2)

        self.bn0 = tf.keras.layers.BatchNormalization()
        self.bn1 = tf.keras.layers.BatchNormalization()

        self.W = tf.Variable(
            initial_value=tf.constant(
                np.random.uniform(-1, 1, (self.rel_vec_dim, self.ent_vec_dim, self.ent_vec_dim)
                                  ), tf.float32),
            trainable=True,
        )

    def call(self, e1_idx, r_idx, training=True):
        e1 = self.E(e1_idx)
        x = self.bn0(e1, training=training)
        x = self.input_dropout(x, training=training)
        x = tf.reshape(x, [-1, 1, e1.shape[1]])

        r = self.R(r_idx)
        W_mat = tf.matmul(r, tf.reshape(self.W, shape=[r.shape[1], -1]))
        W_mat = tf.reshape(W_mat, shape=[-1, e1.shape[1], e1.shape[1]])
        W_mat = self.hidden_dropout1(W_mat, training=training)
        x = tf.matmul(x, W_mat)
        x = tf.reshape(x, shape=[-1, e1.shape[1]])
        x = self.bn1(x, training=training)
        x = self.hidden_dropout2(x, training=training)
        x = tf.matmul(x, tf.transpose(tf.constant(self.E.get_weights()[0]), [1, 0]))
        pred = tf.math.sigmoid(x)
        return pred

    def predict(self, e1_idx, r_idx):
        pre = self(e1_idx, r_idx, training=False)
        return pre

    def get_config(self):
        config = {
            'ent_vec_dim': self.ent_vec_dim,
            "rel_vec_dim": self.rel_vec_dim,
            'input_dropout': self.input_dropout,
            "hidden_dropout1": self.hidden_dropout1,
            "hidden_dropout2": self.hidden_dropout2,
        }
        base_config = super(TuckER, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape



import tensorflow as tf


class CheckpointSaver:
    def __init__(self, model):
        self.checkpoint = tf.train.Checkpoint(model=model)
        self.old_metric_score = 0

    def save(self, filepath,
             metric_score,
             overwrite=True,
             include_optimizer=True,
             save_format=None,
             signatures=None,
             options=None):
        if self.old_metric_score<metric_score:
            path = self.checkpoint.save(filepath, overwrite, include_optimizer,
                                    save_format, signatures, options)
            self.old_metric_score = metric_score
            print("Saving Model in {}".format(path))

